/*
 * SPDX-FileCopyrightText: 2017-2024 Bruce Cowan <bruce@bcowan.me.uk>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "config.h"
#include "rugby-app-window.h"

#include "rugby-list-store.h"
#include "rugby-possibility.h"
#include "rugby-possibility-widget.h"

#include <glib/gi18n.h>

struct _RugbyAppWindow
{
    AdwApplicationWindow parent;

    GtkWidget *scorespin;
    GtkWidget *stack;

    GSettings *win_settings;
};

G_DEFINE_FINAL_TYPE (RugbyAppWindow, rugby_app_window, ADW_TYPE_APPLICATION_WINDOW)

static void
list_store_items_changed_cb (              GListModel     *model,
                             G_GNUC_UNUSED unsigned        position,
                             G_GNUC_UNUSED unsigned        removed,
                             G_GNUC_UNUSED unsigned        added,
                                           RugbyAppWindow *self)
{
    unsigned n_items = g_list_model_get_n_items (model);
    if (n_items == 0)
        gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "empty_page");
    else
        gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "list_page");
}

static char *
item_tooltip_cb (GtkListItem *item)
{
    RugbyPossibility *possibility = gtk_list_item_get_item (item);
    if (!possibility)
      return NULL;

    GString *tooltip = g_string_new (NULL);

    int tries = rugby_possibility_get_tries (possibility);
    int utries = rugby_possibility_get_utries (possibility);
    int kicks = rugby_possibility_get_kicks (possibility);

    if (tries > 0)
    {
        g_string_printf (tooltip, ngettext ("%d converted try",
                                            "%d converted tries",
                                            tries),
                         tries);
    }
    if (utries > 0)
    {
        if (tooltip->len > 0)
            g_string_append (tooltip, ", ");

        g_string_append_printf (tooltip, ngettext ("%d unconverted try",
                                                   "%d unconverted tries",
                                                   utries),
                                utries);
    }
    if (kicks > 0)
    {
        if (tooltip->len > 0)
            g_string_append (tooltip, ", ");

        g_string_append_printf (tooltip, ngettext ("%d kick",
                                                   "%d kicks",
                                                   kicks),
                                  kicks);
    }

    return g_string_free (tooltip, FALSE);
}

static char *
header_label_cb (GtkListItem *item)
{
    RugbyPossibility *possibility = gtk_list_header_get_item (GTK_LIST_HEADER (item));
    if (!possibility)
      return NULL;

    int total_tries = rugby_possibility_total_tries (possibility);

    return g_strdup_printf (ngettext ("%d try", "%d tries", total_tries), total_tries);
}

static void
activate_score_changed (G_GNUC_UNUSED GSimpleAction *action,
                                      GVariant      *parameter,
                                      void          *user_data)
{
    RugbyAppWindow *self = RUGBY_APP_WINDOW (user_data);

    double current_value = gtk_spin_button_get_value (GTK_SPIN_BUTTON (self->scorespin));
    const char *direction = g_variant_get_string (parameter, NULL);

    if (g_strcmp0 (direction, "up") == 0)
        gtk_spin_button_set_value (GTK_SPIN_BUTTON (self->scorespin), current_value + 1.0);
    else if (g_strcmp0 (direction, "down") == 0)
        gtk_spin_button_set_value (GTK_SPIN_BUTTON (self->scorespin), current_value - 1.0);
    else
        g_assert_not_reached ();
}

static void
rugby_app_window_dispose (GObject *object)
{
    gtk_widget_dispose_template (GTK_WIDGET (object), RUGBY_TYPE_APP_WINDOW);
    g_clear_object (&RUGBY_APP_WINDOW (object)->win_settings);

    G_OBJECT_CLASS (rugby_app_window_parent_class)->dispose (object);
}

static void
rugby_app_window_init (RugbyAppWindow *self)
{
    gtk_widget_init_template (GTK_WIDGET (self));

    const GActionEntry win_entries[] = {
        { .name = "score-changed", .activate = activate_score_changed, .parameter_type = "s" },
    };

    g_action_map_add_action_entries (G_ACTION_MAP (self),
                                     win_entries,
                                     G_N_ELEMENTS (win_entries),
                                     self);

    self->win_settings = g_settings_new ("uk.me.bcowan.Rugby.window");
    g_settings_bind (self->win_settings, "height",
                     self, "default-height",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (self->win_settings, "width",
                     self, "default-width",
                     G_SETTINGS_BIND_DEFAULT);
}

static void
rugby_app_window_class_init (RugbyAppWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    g_type_ensure (RUGBY_TYPE_POSSIBILITY_WIDGET);
    g_type_ensure (RUGBY_TYPE_LIST_STORE);

    object_class->dispose = rugby_app_window_dispose;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 "/uk/me/bcowan/Rugby/gtk/window.ui");

    gtk_widget_class_bind_template_child (widget_class, RugbyAppWindow, scorespin);
    gtk_widget_class_bind_template_child (widget_class, RugbyAppWindow, stack);

    gtk_widget_class_bind_template_callback (widget_class, header_label_cb);
    gtk_widget_class_bind_template_callback (widget_class, item_tooltip_cb);
    gtk_widget_class_bind_template_callback (widget_class, list_store_items_changed_cb);
}

RugbyAppWindow *
rugby_app_window_new (GtkApplication *app)
{
    return g_object_new (RUGBY_TYPE_APP_WINDOW,
                         "application", app,
                         NULL);
}
